from unittest.mock import Mock
from crhelper import utils
import unittest
import json


class TestLogHelper(unittest.TestCase):

    def test_send_response(self):
        p = Mock()
        utils._send_response("test_url", {}, put=p)
        p.assert_called_once()
        p = Mock()
        utils._send_response("test_url", p, put=p)
        response = json.loads(p.call_args[1]['data'])
        self.assertEqual("FAILED", response['Status'])
        self.assertEqual({}, response['Data'])



from crhelper.log_helper import *
import unittest
import logging


class TestLogHelper(unittest.TestCase):

    def test_logging_no_formatting(self):
        logger = logging.getLogger('1')
        handler = logging.StreamHandler()
        logger.addHandler(handler)
        orig_formatters = []
        for c in range(len(logging.root.handlers)):
            orig_formatters.append(logging.root.handlers[c].formatter)
        setup(level='DEBUG', formatter_cls=None, boto_level='CRITICAL')
        new_formatters = []
        for c in range(len(logging.root.handlers)):
            new_formatters.append(logging.root.handlers[c].formatter)
        self.assertEqual(orig_formatters, new_formatters)

    def test_logging_boto_explicit(self):
        logger = logging.getLogger('2')
        handler = logging.StreamHandler()
        logger.addHandler(handler)
        setup(level='DEBUG', formatter_cls=None, boto_level='CRITICAL')
        for t in ['boto', 'boto3', 'botocore', 'urllib3']:
            b_logger = logging.getLogger(t)
            self.assertEqual(b_logger.level, 50)

    def test_logging_json(self):
        logger = logging.getLogger('3')
        handler = logging.StreamHandler()
        logger.addHandler(handler)
        setup(level='DEBUG', formatter_cls=JsonFormatter, RequestType='ContainerInit')
        for handler in logging.root.handlers:
            self.assertEqual(JsonFormatter, type(handler.formatter))

    def test_logging_boto_implicit(self):
        logger = logging.getLogger('4')
        handler = logging.StreamHandler()
        logger.addHandler(handler)
        setup(level='DEBUG', formatter_cls=JsonFormatter, RequestType='ContainerInit')
        for t in ['boto', 'boto3', 'botocore', 'urllib3']:
            b_logger = logging.getLogger(t)
            self.assertEqual(b_logger.level, 10)

    def test_logging_json_keys(self):
        with self.assertLogs() as ctx:
            logger = logging.getLogger()
            handler = logging.StreamHandler()
            logger.addHandler(handler)
            setup(level='DEBUG', formatter_cls=JsonFormatter, RequestType='ContainerInit')
            logger.info("test")
            logs = json.loads(ctx.output[0])
        self.assertEqual(["timestamp", "level", "location", "RequestType", "message"], list(logs.keys()))

    def test_logging_json_parse_message(self):
        with self.assertLogs() as ctx:
            logger = logging.getLogger()
            handler = logging.StreamHandler()
            logger.addHandler(handler)
            setup(level='DEBUG', formatter_cls=JsonFormatter, RequestType='ContainerInit')
            logger.info("{}")
            logs = json.loads(ctx.output[0])
        self.assertEqual({}, logs["message"])

    def test_logging_json_exception(self):
        with self.assertLogs() as ctx:
            logger = logging.getLogger()
            handler = logging.StreamHandler()
            logger.addHandler(handler)
            setup(level='DEBUG', formatter_cls=JsonFormatter, RequestType='ContainerInit')
            try:
                1 + 't'
            except Exception as e:
                logger.info("[]", exc_info=True)
            logs = json.loads(ctx.output[0])
        self.assertIn("exception", logs.keys())



from __future__ import print_function
import json
import logging


def _json_formatter(obj):
    """Formatter for unserialisable values."""
    return str(obj)


class JsonFormatter(logging.Formatter):
    """AWS Lambda Logging formatter.

    Formats the log message as a JSON encoded string.  If the message is a
    dict it will be used directly.  If the message can be parsed as JSON, then
    the parse d value is used in the output record.
    """

    def __init__(self, **kwargs):
        super(JsonFormatter, self).__init__()
        self.format_dict = {
            'timestamp': '%(asctime)s',
            'level': '%(levelname)s',
            'location': '%(name)s.%(funcName)s:%(lineno)d',
        }
        self.format_dict.update(kwargs)
        self.default_json_formatter = kwargs.pop(
            'json_default', _json_formatter)

    def format(self, record):
        record_dict = record.__dict__.copy()
        record_dict['asctime'] = self.formatTime(record)

        log_dict = {
            k: v % record_dict
            for k, v in self.format_dict.items()
            if v
        }

        if isinstance(record_dict['msg'], dict):
            log_dict['message'] = record_dict['msg']
        else:
            log_dict['message'] = record.getMessage()

            # Attempt to decode the message as JSON, if so, merge it with the
            # overall message for clarity.
            try:
                log_dict['message'] = json.loads(log_dict['message'])
            except (TypeError, ValueError):
                pass

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            # from logging.Formatter:format
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)

        if record.exc_text:
            log_dict['exception'] = record.exc_text

        json_record = json.dumps(log_dict, default=self.default_json_formatter)

        if hasattr(json_record, 'decode'):  # pragma: no cover
            json_record = json_record.decode('utf-8')

        return json_record


def setup(level='DEBUG', formatter_cls=JsonFormatter, boto_level=None, **kwargs):
    if formatter_cls:
        for handler in logging.root.handlers:
            handler.setFormatter(formatter_cls(**kwargs))

    logging.root.setLevel(level)

    if not boto_level:
        boto_level = level

    logging.getLogger('boto').setLevel(boto_level)
    logging.getLogger('boto3').setLevel(boto_level)
    logging.getLogger('botocore').setLevel(boto_level)
    logging.getLogger('urllib3').setLevel(boto_level)



import crhelper
from test import support
import unittest
from unittest.mock import patch, Mock
import threading

test_events = {
    "Create": {
        "RequestType": "Create",
        "RequestId": "test-event-id",
        "StackId": "arn/test-stack-id/guid",
        "LogicalResourceId": "TestResourceId",
        "ResponseURL": "response_url"
    },
    "Update": {
        "RequestType": "Update",
        "RequestId": "test-event-id",
        "StackId": "test-stack-id",
        "LogicalResourceId": "TestResourceId",
        "PhysicalResourceId": "test-pid",
        "ResponseURL": "response_url"
    },
    "Delete": {
        "RequestType": "Delete",
        "RequestId": "test-event-id",
        "StackId": "test-stack-id",
        "LogicalResourceId": "TestResourceId",
        "PhysicalResourceId": "test-pid",
        "ResponseURL": "response_url"
    }
}


class MockContext(object):

    function_name = "test-function"
    ms_remaining = 9000

    @staticmethod
    def get_remaining_time_in_millis():
        return MockContext.ms_remaining


class TestCfnResource(unittest.TestCase):

    @patch('crhelper.log_helper.setup', return_value=None)
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_init(self, mock_method):
        with support.EnvironmentVarGuard() as environ: 
            environ.set('AWS_REGION', 'us-east-1')
            crhelper.resource_helper.CfnResource()
        mock_method.assert_called_once_with('DEBUG', boto_level='ERROR', formatter_cls=None)

        with support.EnvironmentVarGuard() as environ: 
            environ.set('AWS_REGION', 'us-east-1')
            crhelper.resource_helper.CfnResource(json_logging=True)
        mock_method.assert_called_with('DEBUG', boto_level='ERROR', RequestType='ContainerInit')

    @patch('crhelper.log_helper.setup', return_value=None)
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_init_failure(self, mock_method):
        mock_method.side_effect = Exception("test")
        c = crhelper.resource_helper.CfnResource(json_logging=True)
        self.assertTrue(c._init_failed)

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._polling_init', Mock())
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    @patch('crhelper.resource_helper.CfnResource._wrap_function', Mock())
    @patch('crhelper.resource_helper.CfnResource._cfn_response', return_value=None)
    def test_call(self, cfn_response_mock):
        c = crhelper.resource_helper.CfnResource()
        event = test_events["Create"]
        c.__call__(event, MockContext)
        self.assertTrue(c._send_response)
        cfn_response_mock.assert_called_once_with(event)

        c._sam_local = True
        c._poll_enabled = Mock(return_value=True)
        c._polling_init = Mock()
        c.__call__(event, MockContext)
        c._polling_init.assert_not_called()
        self.assertEqual(1, len(cfn_response_mock.call_args_list))

        c._sam_local = False
        c._send_response = False
        c.__call__(event, MockContext)
        c._polling_init.assert_called()
        self.assertEqual(1, len(cfn_response_mock.call_args_list))

        event = test_events["Delete"]
        c._wait_for_cwlogs = Mock()
        c._poll_enabled = Mock(return_value=False)
        c.__call__(event, MockContext)
        c._wait_for_cwlogs.assert_called()

        c._send = Mock()
        cfn_response_mock.side_effect = Exception("test")
        c.__call__(event, MockContext)
        c._send.assert_called_with('FAILED', "test")

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._polling_init', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    @patch('crhelper.resource_helper.CfnResource._wrap_function', Mock())
    @patch('crhelper.resource_helper.CfnResource._cfn_response', Mock(return_value=None))
    def test_wait_for_cwlogs(self):

        c = crhelper.resource_helper.CfnResource()
        c._context = MockContext
        s = Mock()
        c._wait_for_cwlogs(sleep=s)
        s.assert_not_called()
        MockContext.ms_remaining = 140000
        c._wait_for_cwlogs(sleep=s)
        s.assert_called_once()

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    @patch('crhelper.resource_helper.CfnResource._wrap_function', Mock())
    @patch('crhelper.resource_helper.CfnResource._cfn_response', Mock())
    def test_polling_init(self):
        c = crhelper.resource_helper.CfnResource()
        event = test_events['Create']
        c._setup_polling = Mock()
        c._remove_polling = Mock()
        c._polling_init(event)
        c._setup_polling.assert_called_once()
        c._remove_polling.assert_not_called()
        self.assertEqual(c.PhysicalResourceId, None)

        c.Status = 'FAILED'
        c._setup_polling.assert_called_once()
        c._setup_polling.assert_called_once()

        c = crhelper.resource_helper.CfnResource()
        event = test_events['Create']
        c._setup_polling = Mock()
        c._remove_polling = Mock()
        event['CrHelperPoll'] = "Some stuff"
        c.PhysicalResourceId = None
        c._polling_init(event)
        c._remove_polling.assert_not_called()
        c._setup_polling.assert_not_called()

        c.Status = 'FAILED'
        c._polling_init(event)
        c._remove_polling.assert_called_once()
        c._setup_polling.assert_not_called()

        c.Status = ''
        c.PhysicalResourceId = "some-id"
        c._remove_polling.assert_called()
        c._setup_polling.assert_not_called()

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    @patch('crhelper.resource_helper.CfnResource._wrap_function', Mock())
    def test_cfn_response(self):
        c = crhelper.resource_helper.CfnResource()
        event = test_events['Create']
        c._send = Mock()

        orig_pid = c.PhysicalResourceId
        self.assertEqual(orig_pid, '')
        c._cfn_response(event)
        c._send.assert_called_once()
        print("RID: [%s]" % [c.PhysicalResourceId])
        self.assertEqual(True, c.PhysicalResourceId.startswith('test-stack-id_TestResourceId_'))

        c._send = Mock()
        c.PhysicalResourceId = 'testpid'
        c._cfn_response(event)
        c._send.assert_called_once()
        self.assertEqual('testpid', c.PhysicalResourceId)

        c._send = Mock()
        c.PhysicalResourceId = True
        c._cfn_response(event)
        c._send.assert_called_once()
        self.assertEqual(True, c.PhysicalResourceId.startswith('test-stack-id_TestResourceId_'))

        c._send = Mock()
        c.PhysicalResourceId = ''
        event['PhysicalResourceId'] = 'pid-from-event'
        c._cfn_response(event)
        c._send.assert_called_once()
        self.assertEqual('pid-from-event', c.PhysicalResourceId)

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_wrap_function(self):
        c = crhelper.resource_helper.CfnResource()

        def func(e, c):
            return 'testpid'

        c._wrap_function(func)
        self.assertEqual('testpid', c.PhysicalResourceId)
        self.assertNotEqual('FAILED', c.Status)

        def func(e, c):
            raise Exception('test exception')

        c._wrap_function(func)
        self.assertEqual('FAILED', c.Status)
        self.assertEqual('test exception', c.Reason)

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_send(self):
        c = crhelper.resource_helper.CfnResource()
        s = Mock()
        c._send(send_response=s)
        s.assert_called_once()

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', return_value=None)
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_timeout(self, s):
        c = crhelper.resource_helper.CfnResource()
        c._timeout()
        s.assert_called_with('FAILED', "Execution timed out")

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    def test_set_timeout(self):
        c = crhelper.resource_helper.CfnResource()
        c._context = MockContext()
        def func():
            return None

        c._set_timeout()
        t = threading.Timer(1000, func)
        self.assertEqual(type(t), type(c._timer))
        t.cancel()
        c._timer.cancel()

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_cleanup_response(self):
        c = crhelper.resource_helper.CfnResource()
        c.Data = {"CrHelperPoll": 1, "CrHelperPermission": 2, "CrHelperRule": 3}
        c._cleanup_response()
        self.assertEqual({}, c.Data)

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_remove_polling(self):
        with support.EnvironmentVarGuard() as environ: 
            environ.set('AWS_REGION', 'us-east-1')
            c = crhelper.resource_helper.CfnResource()
        c._context = MockContext()

        c._events_client.remove_targets = Mock()
        c._events_client.delete_rule = Mock()
        c._lambda_client.remove_permission = Mock()

        with self.assertRaises(Exception) as e:
            c._remove_polling()

            self.assertEqual("failed to cleanup CloudWatch event polling", str(e))
        c._events_client.remove_targets.assert_not_called()
        c._events_client.delete_rule.assert_not_called()
        c._lambda_client.remove_permission.assert_not_called()

        c._event["CrHelperRule"] = "1/2"
        c._event["CrHelperPermission"] = "1/2"
        c._remove_polling()
        c._events_client.remove_targets.assert_called()
        c._events_client.delete_rule.assert_called()
        c._lambda_client.remove_permission.assert_called()

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_setup_polling(self):
        with support.EnvironmentVarGuard() as environ: 
            environ.set('AWS_REGION', 'us-east-1')
            c = crhelper.resource_helper.CfnResource()
        c._context = MockContext()
        c._event = test_events["Update"]
        c._lambda_client.add_permission = Mock()
        c._events_client.put_rule = Mock(return_value={"RuleArn": "arn:aws:lambda:blah:blah:function:blah/blah"})
        c._events_client.put_targets = Mock()
        c._setup_polling()
        c._events_client.put_targets.assert_called()
        c._events_client.put_rule.assert_called()
        c._lambda_client.add_permission.assert_called()

    @patch('crhelper.log_helper.setup', Mock())
    @patch('crhelper.resource_helper.CfnResource._poll_enabled', Mock(return_value=False))
    @patch('crhelper.resource_helper.CfnResource._wait_for_cwlogs', Mock())
    @patch('crhelper.resource_helper.CfnResource._send', Mock())
    @patch('crhelper.resource_helper.CfnResource._set_timeout', Mock())
    def test_wrappers(self):
        c = crhelper.resource_helper.CfnResource()

        def func():
            pass

        for f in ["create", "update", "delete", "poll_create", "poll_update", "poll_delete"]:
            self.assertEqual(None, getattr(c, "_%s_func" % f))
            getattr(c, f)(func)
            self.assertEqual(func, getattr(c, "_%s_func" % f))


"""
TODO:
* Async mode – take a wait condition handle as an input, increases max timeout to 12 hours
* Idempotency – If a duplicate request comes in (say there was a network error in signaling back to cfn) the subsequent
  request should return the already created response, will need a persistent store of some kind...
* Functional tests
"""

from __future__ import print_function
import threading
from crhelper.utils import _send_response
from crhelper import log_helper
import logging
import random
import boto3
import string
import json
import os
from time import sleep

logger = logging.getLogger(__name__)

SUCCESS = 'SUCCESS'
FAILED = 'FAILED'


class CfnResource(object):

    def __init__(self, json_logging=False, log_level='DEBUG', boto_level='ERROR', polling_interval=2):
        self._create_func = None
        self._update_func = None
        self._delete_func = None
        self._poll_create_func = None
        self._poll_update_func = None
        self._poll_delete_func = None
        self._timer = None
        self._init_failed = None
        self._json_logging = json_logging
        self._log_level = log_level
        self._boto_level = boto_level
        self._send_response = False
        self._polling_interval = polling_interval
        self.Status = ""
        self.Reason = ""
        self.PhysicalResourceId = ""
        self.StackId = ""
        self.RequestId = ""
        self.LogicalResourceId = ""
        self.Data = {}
        self._event = {}
        self._context = None
        self._response_url = ""
        self._sam_local = os.getenv('AWS_SAM_LOCAL')
        self._region = os.getenv('AWS_REGION')
        try:
            if not self._sam_local:
                self._lambda_client = boto3.client('lambda', region_name=self._region)
                self._events_client = boto3.client('events', region_name=self._region)
                self._logs_client = boto3.client('logs', region_name=self._region)
            if json_logging:
                log_helper.setup(log_level, boto_level=boto_level, RequestType='ContainerInit')
            else:
                log_helper.setup(log_level, formatter_cls=None, boto_level=boto_level)
        except Exception as e:
            logger.error(e, exc_info=True)
            self.init_failure(e)

    def __call__(self, event, context):
        try:
            self._log_setup(event, context)
            logger.debug(event)
            self._crhelper_init(event, context)
            # Check for polling functions
            if self._poll_enabled() and self._sam_local:
                logger.info("Skipping poller functionality, as this is a local invocation")
            elif self._poll_enabled():
                self._polling_init(event)
            # If polling is not enabled, then we should respond
            else:
                logger.debug("enabling send_response")
                self._send_response = True
            logger.debug("_send_response: %s" % self._send_response)
            if self._send_response:
                if self.RequestType == 'Delete':
                    self._wait_for_cwlogs()
                self._cfn_response(event)
        except Exception as e:
            logger.error(e, exc_info=True)
            self._send(FAILED, str(e))
        finally:
            if self._timer:
                self._timer.cancel()

    def _wait_for_cwlogs(self, sleep=sleep):
        sleep_time = int(self._context.get_remaining_time_in_millis() / 1000) - 15
        if sleep_time > 120:
            sleep_time = 120
        if sleep_time > 1:
            sleep(sleep_time)

    def _log_setup(self, event, context):
        if self._json_logging:
            log_helper.setup(self._log_level, boto_level=self._boto_level, RequestType=event['RequestType'],
                             StackId=event['StackId'], RequestId=event['RequestId'],
                             LogicalResourceId=event['LogicalResourceId'], aws_request_id=context.aws_request_id)
        else:
            log_helper.setup(self._log_level, boto_level=self._boto_level, formatter_cls=None)

    def _crhelper_init(self, event, context):
        self._send_response = False
        self.Status = SUCCESS
        self.Reason = ""
        self.PhysicalResourceId = ""
        self.StackId = event["StackId"]
        self.RequestId = event["RequestId"]
        self.LogicalResourceId = event["LogicalResourceId"]
        self.Data = {}
        if "CrHelperData" in event.keys():
            self.Data = event["CrHelperData"]
        self.RequestType = event["RequestType"]
        self._event = event
        self._context = context
        self._response_url = event['ResponseURL']
        if self._timer:
            self._timer.cancel()
        if self._init_failed:
            return self._send(FAILED, str(self._init_failed))
        self._set_timeout()
        self._wrap_function(self._get_func())

    def _polling_init(self, event):
        # Setup polling on initial request
        logger.debug("pid1: %s" % self.PhysicalResourceId)
        if 'CrHelperPoll' not in event.keys() and self.Status != FAILED:
            logger.info("Setting up polling")
            self.Data["PhysicalResourceId"] = self.PhysicalResourceId
            self._setup_polling()
            self.PhysicalResourceId = None
            logger.debug("pid2: %s" % self.PhysicalResourceId)
        # if physical id is set, or there was a failure then we're done
        logger.debug("pid3: %s" % self.PhysicalResourceId)
        if self.PhysicalResourceId or self.Status == FAILED:
            logger.info("Polling complete, removing cwe schedule")
            self._remove_polling()
            self._send_response = True

    def _cfn_response(self, event):
        # Use existing PhysicalResourceId if it's in the event and no ID was set
        if not self.PhysicalResourceId and "PhysicalResourceId" in event.keys():
            logger.info("PhysicalResourceId present in event, Using that for response")
            self.PhysicalResourceId = event['PhysicalResourceId']
        # Generate a physical id if none is provided
        elif not self.PhysicalResourceId or self.PhysicalResourceId is True:
            if "PhysicalResourceId" in event.keys():
                logger.info("PhysicalResourceId present in event, Using that for response")
            logger.info("No physical resource id returned, generating one...")
            self.PhysicalResourceId = event['StackId'].split('/')[1] + '_' + event[
                'LogicalResourceId'] + '_' + self._rand_string(8)
        self._send()

    def _poll_enabled(self):
        return getattr(self, "_poll_{}_func".format(self._event['RequestType'].lower()))

    def create(self, func):
        self._create_func = func
        return func

    def update(self, func):
        self._update_func = func
        return func

    def delete(self, func):
        self._delete_func = func
        return func

    def poll_create(self, func):
        self._poll_create_func = func
        return func

    def poll_update(self, func):
        self._poll_update_func = func
        return func

    def poll_delete(self, func):
        self._poll_delete_func = func
        return func

    def _wrap_function(self, func):
        try:
            self.PhysicalResourceId = func(self._event, self._context)
        except Exception as e:
            logger.error(str(e), exc_info=True)
            self.Reason = str(e)
            self.Status = FAILED

    def _timeout(self):
        logger.error("Execution is about to time out, sending failure message")
        self._send(FAILED, "Execution timed out")

    def _set_timeout(self):
        self._timer = threading.Timer((self._context.get_remaining_time_in_millis() / 1000.00) - 0.5,
                                      self._timeout)
        self._timer.start()

    def _get_func(self):
        request_type = "_{}_func"
        if "CrHelperPoll" in self._event.keys():
            request_type = "_poll" + request_type
        return getattr(self, request_type.format(self._event['RequestType'].lower()))

    def _send(self, status=None, reason="", send_response=_send_response):
        if len(str(str(self.Reason))) > 256:
            self.Reason = "ERROR: (truncated) " + str(self.Reason)[len(str(self.Reason)) - 240:]
        if len(str(reason)) > 256:
            reason = "ERROR: (truncated) " + str(reason)[len(str(reason)) - 240:]
        response_body = {
            'Status': self.Status,
            'PhysicalResourceId': str(self.PhysicalResourceId),
            'StackId': self.StackId,
            'RequestId': self.RequestId,
            'LogicalResourceId': self.LogicalResourceId,
            'Reason': str(self.Reason),
            'Data': self.Data,
        }
        if status:
            response_body.update({'Status': status, 'Reason': reason})
        send_response(self._response_url, response_body)

    def init_failure(self, error):
        self._init_failed = error
        logger.error(str(error), exc_info=True)

    def _cleanup_response(self):
        for k in ["CrHelperPoll", "CrHelperPermission", "CrHelperRule"]:
            if k in self.Data.keys():
                del self.Data[k]

    @staticmethod
    def _rand_string(l):
        return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(l))

    def _add_permission(self, rule_arn):
        sid = self._event['LogicalResourceId'] + self._rand_string(8)
        self._lambda_client.add_permission(
            FunctionName=self._context.function_name,
            StatementId=sid,
            Action='lambda:InvokeFunction',
            Principal='events.amazonaws.com',
            SourceArn=rule_arn
        )
        return sid

    def _put_rule(self):
        response = self._events_client.put_rule(
            Name=self._event['LogicalResourceId'] + self._rand_string(8),
            ScheduleExpression='rate({} minutes)'.format(self._polling_interval),
            State='ENABLED',
        )
        return response["RuleArn"]

    def _put_targets(self, func_name):
        region = self._event['CrHelperRule'].split(":")[3]
        account_id = self._event['CrHelperRule'].split(":")[4]
        partition = self._event['CrHelperRule'].split(":")[1]
        rule_name = self._event['CrHelperRule'].split("/")[1]
        logger.debug(self._event)
        self._events_client.put_targets(
            Rule=rule_name,
            Targets=[
                {
                    'Id': '1',
                    'Arn': 'arn:%s:lambda:%s:%s:function:%s' % (partition, region, account_id, func_name),
                    'Input': json.dumps(self._event)
                }
            ]
        )

    def _remove_targets(self, rule_arn):
        self._events_client.remove_targets(
            Rule=rule_arn.split("/")[1],
            Ids=['1']
        )

    def _remove_permission(self, sid):
        self._lambda_client.remove_permission(
            FunctionName=self._context.function_name,
            StatementId=sid
        )

    def _delete_rule(self, rule_arn):
        self._events_client.delete_rule(
            Name=rule_arn.split("/")[1]
        )

    def _setup_polling(self):
        self._event['CrHelperData'] = self.Data
        self._event['CrHelperPoll'] = True
        self._event['CrHelperRule'] = self._put_rule()
        self._event['CrHelperPermission'] = self._add_permission(self._event['CrHelperRule'])
        self._put_targets(self._context.function_name)

    def _remove_polling(self):
        if 'CrHelperData' in self._event.keys():
            self._event.pop('CrHelperData')
        if "PhysicalResourceId" in self.Data.keys():
            self.Data.pop("PhysicalResourceId")
        if 'CrHelperRule' in self._event.keys():
            self._remove_targets(self._event['CrHelperRule'])
        else:
            logger.error("Cannot remove CloudWatch events rule, Rule arn not available in event")
        if 'CrHelperPermission' in self._event.keys():
            self._remove_permission(self._event['CrHelperPermission'])
        else:
            logger.error("Cannot remove lambda events permission, permission id not available in event")
        if 'CrHelperRule' in self._event.keys():
            self._delete_rule(self._event['CrHelperRule'])
        else:
            logger.error("Cannot remove CloudWatch events target, Rule arn not available in event")



from __future__ import print_function
from botocore.vendored import requests
import json
import logging as logging
import time

logger = logging.getLogger(__name__)


def _send_response(response_url, response_body, put=requests.put):
    try:
        json_response_body = json.dumps(response_body)
    except Exception as e:
        msg = "Failed to convert response to json: {}".format(str(e))
        logger.error(msg, exc_info=True)
        response_body = {'Status': 'FAILED', 'Data': {}, 'Reason': msg}
        json_response_body = json.dumps(response_body)
    logger.debug("CFN response URL: {}".format(response_url))
    logger.debug(json_response_body)
    headers = {'content-type': '', 'content-length': str(len(json_response_body))}
    while True:
        try:
            response = put(response_url, data=json_response_body, headers=headers)
            logger.info("CloudFormation returned status code: {}".format(response.reason))
            break
        except Exception as e:
            logger.error("Unexpected failure sending response to CloudFormation {}".format(e), exc_info=True)
            time.sleep(5)




"""
Copyright 2016-2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the
License. A copy of the License is located at
    http://aws.amazon.com/apache2.0/
or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and
limitations under the License.
"""

from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="crhelper",
    version="2.0.3",
    description="crhelper simplifies authoring CloudFormation Custom Resources",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aws-cloudformation/custom-resource-helper",
    author="Jay McConnell",
    author_email="jmmccon@amazon.com",
    license="Apache2",
    packages=find_packages(),
    tests_require=["boto3"],
    test_suite="tests",
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
